@@ -1,8 +1,8 @@
 import logging
 import os
+import re
 from importlib import import_module
 from importlib.util import module_from_spec, spec_from_file_location
-from uuid import uuid4
 
 from bs4 import BeautifulSoup, PageElement
 from weasyprint import HTML, urls
@@ -13,6 +13,10 @@ from .preprocessor import get_combined as prep_combined
 from .styles import style_for_print
 from .themes import generic as generic_theme
 from .toc import make_indexes
+from .utils.emoji_util import fix_twemoji
+from .utils.image_util import fix_image_alignment
+from .utils.iframe_util import convert_iframe
+from .utils.section import get_section_path
 from .utils.soup_util import clone_element
 
 
@@ -66,15 +70,14 @@ class Generator(object):
 
         if article:
             # remove 'headerlink' if exists.
-            for a in article.find_all('a', attrs={'class': 'headerlink'}):
+            for a in article.select('a.headerlink'):
                 a.decompose()
-
-            if self._options.heading_shift:
-                article = self._shift_heading(article, page, page.parent)
-
-            article['id'] = f'{page.url}:'  # anchor for each page.
-            article['data-url'] = f'/{page.url}'
+            for a in article.select('a.md-content__button'):
+                a.decompose()
+            self._fix_missing_id_for_h1(article, page)
             setattr(page, 'pdf-article', article)
+        else:
+            self.logger.warning(f'Missing article: [{page.title}]({page.url})')
 
         return self._theme.inject_link(output_content, pdf_path)
 
@@ -95,12 +98,23 @@ class Generator(object):
         add_stylesheet(self._theme.get_stylesheet())
 
         for page in self._nav:
-            self._add_content(soup, page)
+            content = self._get_content(soup, page)
+            if content:
+                soup.body.append(content)
 
         make_indexes(soup, self._options)
         make_cover(soup, self._options)
 
+        fix_image_alignment(soup, self._options.logger)
+        fix_twemoji(soup, self._options.logger)
+
+        if len(self._options.convert_iframe) > 0:
+            convert_iframe(soup,
+                           self._options.convert_iframe,
+                           self._options.logger)
+
         if self._options.debug_html:
+            self._link_check(soup)
             print(f'{soup}')
 
         html = HTML(string=str(soup))
@@ -109,6 +123,7 @@ class Generator(object):
         abs_pdf_path = os.path.join(config['site_dir'], output_path)
         os.makedirs(os.path.dirname(abs_pdf_path), exist_ok=True)
 
+        self.logger.info(f'Output a PDF to "{abs_pdf_path}".')
         render.write_pdf(abs_pdf_path)
 
     # ------------------------
@@ -133,30 +148,20 @@ class Generator(object):
             if not hit:
                 break
 
-    def _shift_heading(self, article: PageElement, page, parent):
-        if not parent:
-            return article
+    def _page_path_for_id(self, page):
+        """ normalize to directory urls style"""
 
-        for i in range(7, 0, -1):
-            while True:
-                h = article.find(f'h{i}')
-                if not h:
-                    break
-                h.name = f'h{i + 1}'
+        if page.is_section:
+            path = get_section_path(page)
+        else:
+            path = page.url
 
-        if page == parent.children[0]:
-            article = self._shift_heading(article, parent, parent.parent)
-
-            id = page.url if (hasattr(page, 'url') and page.url) else ''
-            id = id.strip('/').rstrip('/')
-            id = id if id else str(uuid4())
-
-            soup = BeautifulSoup('', 'html.parser')
-            h1 = soup.new_tag('h1', id=f'{id}')
-            h1.append(parent.title)
-            article.insert(0, h1)
-
-        return article
+        path = '.' if not path or path == 'index.html' else path
+        if path.endswith('index.html'):
+            path = re.sub(r'index\.html$', '', path)
+        elif path.endswith('.html'):
+            path = re.sub(r'\.html$', '/', path)
+        return path
 
     def _soup_from_content(self, content: str, page):
         soup = BeautifulSoup(content, 'html.parser')
@@ -175,13 +180,73 @@ class Generator(object):
 
         return prep_combined(soup, base_url, page.file.url)
 
-    def _add_content(self, soup: PageElement, page):
+    def _get_content(self, soup: PageElement, page):
+
+        def shift_heading(elem, page):
+            for i in range(7, 0, -1):
+                while True:
+                    h = elem.find(f'h{i}')
+                    if not h:
+                        break
+                    h.name = f'h{i + 1}'
+
+            page_path = self._page_path_for_id(page)
+            h1 = soup.new_tag('h1', id=f'{page_path}')
+            h1.append(page.title)
+            elem.insert(0, h1)
+            return elem
+
+        def cleanup_class(classes: []):
+            if classes and len(classes):
+                excludes = ['md-content__inner']
+                return [c for c in classes if not (c in excludes)]
+            return classes
+
         article = getattr(page, 'pdf-article', None)
         if article:
-            soup.body.append(article)
-        if page.children:
+
+            page_path = self._page_path_for_id(page)
+            article['id'] = f'{page_path}:'  # anchor for each page.
+            article['data-url'] = f'/{page_path}'
+            return article
+
+        elif page.children:
+
+            new_article = soup.new_tag('article')
+            found = False
             for c in page.children:
-                self._add_content(soup, c)
+                content = self._get_content(soup, c)
+                if content:
+                    new_article.append(content)
+                    found = True
+
+            if not found:
+                return None
+
+            child_classes = None
+            for child_article in new_article.find_all('article'):
+                child_article.name = 'section'
+                classes = child_article.get('class')
+                if classes and not child_classes:
+                    child_classes = classes
+                child_article['class'] = cleanup_class(classes)
+
+            page_path = self._page_path_for_id(page)
+            new_article['id'] = f'{page_path}:'  # anchor for each page.
+            new_article['data-url'] = f'/{page_path}'
+            if child_classes:
+                new_article['class'] = child_classes
+
+            if self._options.heading_shift:
+                return shift_heading(new_article, page)
+            return new_article
+
+        return None
+
+    def _fix_missing_id_for_h1(self, content, page):
+        h1 = content.find('h1')
+        if h1 and not h1.get('id'):
+            h1['id'] = self._page_path_for_id(page)
 
     # -------------------------------------------------------------
 
@@ -212,3 +277,35 @@ class Generator(object):
         except ImportError as e:
             self.logger.error(f'Could not load theme handler {theme}: {e}')
             return generic_theme
+
+    # -------------------------------------------------------------
+
+    def _link_check(self, soup):
+        from urllib.parse import urlparse
+
+        anchors = set(map(lambda el: '#' + el['id'], soup.find_all(id=True)))
+
+        '''
+        from pprint import pprint as pp
+        pp(anchors)
+        '''
+
+        missing = set()
+        for el in soup.find_all('a', href=True):
+            href = el['href']
+            target_url = urlparse(href)
+
+            if target_url.scheme or target_url.netloc:
+                continue
+            if href in anchors:
+                continue
+
+            missing.add(href)
+
+        if len(missing):
+            self.logger.error(f'Missing {len(missing)} link(s):')
+            for link in sorted(missing):
+                self.logger.warning(f'  | {link}')
+            self.logger.info('  | --- found anchors:')
+            for anchor in sorted(anchors):
+                self.logger.info(f'  | {anchor}')
