@@ -28,12 +28,14 @@ class Options(object):
         ('cover_title', config_options.Type(str, default=None)),
         ('cover_subtitle', config_options.Type(str, default=None)),
 
+        ('toc_title', config_options.Type(str, default="Table of contents")),
         ('heading_shift', config_options.Type(bool, default=True)),
         ('toc_level', config_options.Type(int, default=2)),
         ('ordered_chapter_level', config_options.Type(int, default=3)),
         ('excludes_children', config_options.Type(list, default=[])),
 
-        ('exclude_pages', config_options.Type(list, default=[]))
+        ('exclude_pages', config_options.Type(list, default=[])),
+        ('convert_iframe', config_options.Type(list, default=[]))
     )
 
     def __init__(self, local_config, config, logger: logging):
@@ -60,6 +62,7 @@ class Options(object):
             self._cover_subtitle = local_config['cover_subtitle']
 
         # TOC and Chapter heading
+        self.toc_title = _normalize(local_config['toc_title'])
         self.heading_shift = local_config['heading_shift']
         self.toc_level = local_config['toc_level']
         self.ordered_chapter_level = local_config['ordered_chapter_level']
@@ -67,6 +70,7 @@ class Options(object):
 
         # Page
         self.exclude_pages = local_config['exclude_pages']
+        self.convert_iframe = local_config['convert_iframe']
 
         # Theming
         self.theme_name = config['theme'].name
